#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2017, California Institute of Technology
# All rights reserved.

g_program_name = __file__.split('/')[-1]   # = 'charge_by_bond.py'
g_date_str = '2020-11-04'
g_version_str = '0.15.0'


import sys
import re
from collections import defaultdict
import re

try:
    from . import ttree_lex
    from .lttree_styles import AtomStyle2ColNames, ColNames2AidAtypeMolid
except (ImportError, SystemError, ValueError):
    # not installed as a package
    import ttree_lex
    from lttree_styles import AtomStyle2ColNames, ColNames2AidAtypeMolid



def LookupChargePairs(chargebyatomid,
                      # bond_ids,
                      # bond_types,
                      # bond_pairs,
                      lines_atoms,
                      lines_bonds,
                      lines_bond_list,
                      lines_chargebybond,
                      atom_style,
                      section_name,
                      prefix='',
                      suffix=''):
                      # bond_ids_offset=0):
                      # report_progress = False):
    """
    LookupChargePairs() looks up partial-charge pair contributions from the
    types of atoms participating in a bond.

    Output:
    ...It looks up the corresponding change in the partial charges for
       each pair of atoms and stores this in the "chargebyatomid" dictionary.

    Input (continued):
       This function requires:
    ...a list of bonded pairs of atoms
        stored in the lines_bonds variable (from the "Data Bond List"
                                       or "Data Bonds AtomId AtomId" sections)
    ...and a list of atom types
        stored in the lines_atoms variable (from the "Data Atoms" section)

    ...and a list of charge-pairs-as-a-function-of-atom-types
        stored in the lines_chargebybond (from the "Data Bonds By Type" section)

    """

    column_names = AtomStyle2ColNames(atom_style)
    i_atomid, i_atomtype, i_molid = ColNames2AidAtypeMolid(column_names)

    atomids = []
    atomtypes = []
    atomids2types = {}

    for iv in range(0, len(lines_atoms)):
        line = lines_atoms[iv].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) > 0:
            tokens = ttree_lex.SplitQuotedString(line)
            if ((len(tokens) <= i_atomid) or (len(tokens) <= i_atomtype)):
                sys.stderr.write("\"" + line + "\"\n")
                raise(ttree_lex.InputError(
                    'Error not enough columns on line ' + str(iv + 1) + ' of \"Atoms\" section.'))
            tokens = ttree_lex.SplitQuotedString(line)
            atomid = ttree_lex.EscCharStrToChar(tokens[i_atomid])
            atomids.append(atomid)
            atomtype = ttree_lex.EscCharStrToChar(tokens[i_atomtype])
            atomtypes.append(atomtype)
            atomids2types[atomid] = atomtype

    #assert(isinstance(bond_ids, list))
    #assert(isinstance(bond_types, list))
    #assert(isinstance(bond_pairs, list))
    #del bond_ids[:]
    #del bond_types[:]
    #del bond_pairs[:]
    bond_pairs = []

    for ie in range(0, len(lines_bond_list)):
        line = lines_bond_list[ie].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) == 0:
            continue
        tokens = ttree_lex.SplitQuotedString(line)
        if len(tokens) == 3:
            # bond_ids.append(ttree_lex.EscCharStrToChar(tokens[0]))
            bond_pairs.append((ttree_lex.EscCharStrToChar(tokens[1]),
                               ttree_lex.EscCharStrToChar(tokens[2])))
        else:
            raise(ttree_lex.InputError('Incorrect number of columns on line ' +
                                       str(ie + 1) + ' of \"' + section_name + '\" section.'))

    for ie in range(0, len(lines_bonds)):
        line = lines_bonds[ie].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) == 0:
            continue
        tokens = ttree_lex.SplitQuotedString(line)
        if len(tokens) == 4:
            # bond_ids.append(ttree_lex.EscCharStrToChar(tokens[0]))
            # bond_types.append(ttree_lex.EscCharStrToChar(tokens[1]))
            bond_pairs.append((ttree_lex.EscCharStrToChar(tokens[2]),
                               ttree_lex.EscCharStrToChar(tokens[3])))
        else:
            raise(ttree_lex.InputError('Incorrect number of columns on line ' +
                                       str(ie + 1) + ' of \"' + section_name + '\" section.'))

    # for ie in range(0, len(lines_bonds_atomid_atomid)):
    #    line = lines_bonds_atomid_atomid[ie].strip()
    #    if '#' in line:
    #        icomment = line.find('#')
    #        line = (line[:icomment]).strip()
    #    if len(line) == 0:
    #        continue
    #    tokens = ttree_lex.SplitQuotedString(line)
    #    if len(tokens) == 2:
    #        #bondid_n = bond_ids_offset + len(bond_ids) + 1
    #        #bond_ids.append(prefix+str(bondid_n)+suffix)
    #        bond_pairs.append( (ttree_lex.EscCharStrToChar(tokens[0]),
    #                            ttree_lex.EscCharStrToChar(tokens[1])) )
    #    else:
    #        raise(ttree_lex.InputError('Incorrect number of columns on line '+str(ie+1)+' of \"'+section_name+'\" section.'))

    #assert(len(bond_types) == 0)
    typepattern_to_chargepairs = []
    warning_unassigned_chargepairs = None

    for i in range(0, len(lines_chargebybond)):
        line = lines_chargebybond[i].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) > 0:
            tokens = ttree_lex.SplitQuotedString(line)

            if (len(tokens) != 4):
                raise(ttree_lex.InputError('Error: Wrong number of columns in the \"Charge Pairs By Type\" section of data file.\n'
                                           'Offending line:\n' +
                                           '\"' + line + '\"\n'
                                           'Expected 4 columns\n'))

            chargepair = (float(tokens[2]),
                          float(tokens[3]))

            typepattern = []

            for typestr in tokens[:2]:
                if ttree_lex.HasRE(typestr):
                    regex_str = VarNameToRegex(typestr)
                    typepattern.append(re.compile(regex_str))
                else:
                    typepattern.append(ttree_lex.EscCharStrToChar(typestr))

            typepattern_to_chargepairs.append([typepattern, chargepair])

    for atomid1, atomid2 in bond_pairs:

        if atomid1 not in atomids2types:
            raise ttree_lex.InputError('Error: atom \"' + atomid1 + '\" not defined in \"Data Atoms\".\n'
                                       '       This usually happens when the user mistypes one of the names of the\n'
                                       '       $atoms in either a \"Data Atoms\" or \"Data Bond List\" section.\n'
                                       '       To find out where the mistake occured, search the \n'
                                       '       \"ttree_assignments.txt\" file for:\n'
                                       '       \"' + atomid1 + '\"\n')

        if atomid2 not in atomids2types:
            raise ttree_lex.InputError('Error: atom \"' + atomid2 + '\" not defined in \"Data Atoms\".\n'
                                       '       This usually happens when the user mistypes one of the names of the\n'
                                       '       $atoms in either a \"Data Atoms\" or \"Data Bond List\" section.\n'
                                       '       To find out where the mistake occured, search the \n'
                                       '       \"ttree_assignments.txt\" file for:\n'
                                       '       \"' + atomid2 + '\"\n')

        atomtype1 = atomids2types[atomid1]
        atomtype2 = atomids2types[atomid2]

        found = False
        for typepattern, chargepair in typepattern_to_chargepairs:
            # use string comparisons to check if atom types match the pattern
            if ttree_lex.MatchesAll((atomtype1, atomtype2), typepattern):
                # ("MatchesAll()" defined in "ttree_lex.py")
                chargebyatomid[atomid1] += chargepair[0]
                chargebyatomid[atomid2] += chargepair[1]
                found = True
            elif ttree_lex.MatchesAll((atomtype2, atomtype1), typepattern):
                chargebyatomid[atomid1] += chargepair[1]
                chargebyatomid[atomid2] += chargepair[0]
                found = True
        if (not found) and (not warning_unassigned_chargepairs):
            warning_unassigned_chargepairs = (atomid1, atomid2)

    if warning_unassigned_chargepairs:
        sys.stderr.write('---------------------------------------------------------------------------\n'
                         'Warning: bonds found between atoms with no partial-charge rules.\n'
                         '         This means that somewhere you are using a force-field\n'
                         '         which assigns atomic charge according to the bonds these atoms\n'
                         '         participate in, AND at least one pair of bonded atoms does NOT have\n'
                         '         a rule defined to assign charges to that pair of atoms.\n'
                         '         This can happen if there is a problem with the force-field file\n'
                         '         OR if you are defining the charges for these atoms manually\n'
                         '         In the later case, it is not a problem.\n'
                         '         The first bond with this problem is between this pair of atoms:\n'
                         '            ' +
                         str(warning_unassigned_chargepairs[0]) + '\n'
                         '            ' +
                         str(warning_unassigned_chargepairs[1]) + '\n'
                         '---------------------------------------------------------------------------\n')

def main():
    """
    This is is a "main module" wrapper for invoking chargepairs_by_type.py
    as a stand alone program.  This program:
    This program reads a LAMMPS data file (or an excerpt of a LAMMPS)
    data file containing bonded many-body interactions by atom type
    (and bond type), and generates a list of atom charges in LAMMPS input
    script format consistent with those types (to the standard out).

    Typical Usage:

    chargepairs_by_type.py -atoms atoms.data \\
                           -bonds bonds.data \\
                           -chargebybond chargepairs_by_type.data \\
                           > list_of_atom_charges.in

    """

    #######  Main Code Below: #######
    sys.stderr.write(g_program_name + ' v' +
                     g_version_str + ' ' + g_date_str + ' ')
    if sys.version < '3':
        sys.stderr.write(' (python version < 3)\n')
    else:
        sys.stderr.write('\n')

    try:
        fname_atoms = None
        fname_bonds = None
        fname_bond_list = None
        fname_chargebybond = None
        section_name = 'Data Bond List'  # (This will be replaced later.)
        atom_style = 'full'
        prefix = ''
        suffix = ''
        bond_lack_types = False

        argv = [arg for arg in sys.argv]

        # Loop over the remaining arguments not processed yet.
        # These arguments are specific to the lttree.py program
        # and are not understood by ttree.py:
        i = 1
        while i < len(argv):
            #sys.stderr.write('argv['+str(i)+'] = \"'+argv[i]+'\"\n')
            if ((argv[i].lower() == '-?') or
                    (argv[i].lower() == '--?') or
                    (argv[i].lower() == '-help') or
                    (argv[i].lower() == '-help')):
                if i + 1 >= len(argv):
                    sys.stdout.write(man_page_text + '\n')
                    sys.exit(0)

            elif argv[i].lower() == '-atoms':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError('Error: ' + argv[i] + ' flag should be followed by a file name containing lines of\n'
                                               '       text which might appear in the "Atoms" section of a LAMMPS data file.\n')
                fname_atoms = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-bonds':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError('Error: ' + argv[i] + ' flag should be followed by a file name containing lines of\n'
                                               '       text which might appear in the "Bonds" section of a LAMMPS data file.\n')
                fname_bonds = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-bond-list':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError(
                        'Error: ' + argv[i] + ' flag should be followed by a file name\n')
                    # raise ttree_lex.InputError('Error: '+argv[i]+' flag should be followed by a file name containing lines of\n'
                    #                 '       text which might appear in the "Bonds No Types" section of a LAMMPS data file.\n')
                fname_bond_list = argv[i + 1]
                section_name = "Data Bond List"
                del(argv[i:i + 2])

            elif ((argv[i].lower() == '-chargebybond') or
                  (argv[i].lower() == '-chargesbybond') or
                  (argv[i].lower() == '-charge-by-bond') or
                  (argv[i].lower() == '-charges-by-bond') or
                  (argv[i].lower() == '-chargepairsbytype') or
                  (argv[i].lower() == '-chargepairs-by-type') or
                  (argv[i].lower() == '-charge-pairs-by-type')):
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError(
                        'Error: ' + argv[i] + ' flag should be followed by a file name\n')

                    # raise ttree_lex.InputError('Error: '+argv[i]+' flag should be followed by a file name containing\n'
                    #                 '       text which might appear in the "'+section_name+' By Type" section\n'
                    #                 '       of a LAMMPS data file.\n')
                fname_chargebybond = argv[i + 1]
                del(argv[i:i + 2])

            elif ((argv[i].lower() == '-atom-style') or
                  (argv[i].lower() == '-atom_style')):
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError('Error: ' + argv[i] + ' flag should be followed by a an atom_style name.\n'
                                               '       (Or single quoted string which includes a space-separated\n'
                                               '       list of column names.)\n')
                atom_style = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i][0] == '-':
                raise ttree_lex.InputError('Error(' + g_program_name + '):\n'
                                           'Unrecogized command line argument \"' + argv[i] + '\"\n')
            else:
                i += 1

        if len(argv) != 1:
            # if there are more than 2 remaining arguments,
            problem_args = ['\"' + arg + '\"' for arg in argv[1:]]
            raise ttree_lex.InputError('Syntax Error(' + g_program_name + '):\n\n'
                                       '       Problem with argument list.\n'
                                       '       The remaining arguments are:\n\n'
                                       '         ' +
                                       (' '.join(problem_args)) + '\n\n'
                                       '       (The actual problem may be earlier in the argument list.)\n')

        #bond_types = []
        fatoms = open(fname_atoms, 'r')
        lines_bonds = []
        lines_bond_list = []
        fbonds = fbond_list = None
        try:
            if fname_bonds != None:
                fbonds = open(fname_bonds, 'r')
                lines_bonds = fbonds.readlines()
                fbonds.close()
        except IOError:
            pass
        try:
            if fname_bond_list != None:
                fbond_list = open(fname_bond_list, 'r')
                lines_bond_list = fbond_list.readlines()
                fbond_list.close()
        except IOError:
            pass
        if ((len(lines_bonds) == 0) and (len(lines_bond_list) == 0)):
            sys.stderr.write('Error(' + g_program_name + '): No bonds defined for this system\n'
                             '      (This error may be a bug in moltemplate.)\n')
        fchargebybond = open(fname_chargebybond, 'r')
        lines_atoms = fatoms.readlines()

        lines_chargebybond = fchargebybond.readlines()
        fatoms.close()
        fchargebybond.close()
        chargebyatomid = defaultdict(float)

        LookupChargePairs(chargebyatomid,
                          lines_atoms,
                          lines_bonds,
                          lines_bond_list,
                          lines_chargebybond,
                          atom_style,
                          section_name)

        for atomid, charge in chargebyatomid.items():
            sys.stdout.write('  set atom ' + str(atomid) +
                             ' charge ' + str(charge) + '\n')

    except (ValueError, ttree_lex.InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(-1)



if __name__ == "__main__":
    main()




try:
    from ..nbody_graph_search import Ugraph
except (SystemError, ValueError):
    # not installed as a package
    from nbody_graph_search import Ugraph


# This file defines how 3-body angle interactions are generated by moltemplate
# by default.  It can be overridden by supplying your own custom file.

#    To find 3-body "angle" interactions, we would use this subgraph:
#
#
#       *---*---*           =>  1st bond connects atoms 0 and 1
#       0   1   2               2nd bond connects atoms 1 and 2
#

bond_pattern = Ugraph([(0, 1), (1, 2)])
# (Ugraph atom indices begin at 0, not 1)


#    The next function eliminates the redundancy between 0-1-2 and 2-1-0:
def canonical_order(match):
    """
    When searching for atoms with matching bond patterns GraphMatcher
    often returns redundant results. We must define a "canonical_order"
    function which sorts the atoms and bonds in a way which is consistent
    with the type of N-body interaction being considered.
    However, some angle_styles (such as angle_style class2)
    have no symmetry (at least not for arbitrary choices of parameters).
    These force-field styles, the different permulations of atom-order
    are not equivalent.  So we do not want to rearrange the order of
    the atoms (and bonds) in the match, because the formula for the
    interaction between atoms 1,2,3 is not the same as the formula
    for the interaction between atoms 3,2,1.
    In this case, this function returns
    the original "match" argument unmodified.

    """

    return match



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013

"""
   Unfortunately, the lines in a PDB files are not always listed in the
   correct order.  Software that reads PDB files is expected to re-sort this
   data before interpreting it.  (One reason I don't like them.)
   This script reads a PDB file from the standard input, sorts the lines
   according to ChainID, SeqNum (residue-number), Icode (insert-code),
   and AtomID (in that order) and prints the result to the standard-out.
   Only the ATOM and HETATM records are effected.
   All other lines in the PDB file are printed back to the user verbatim.
   Note: the "altLoc" column (character #17 on each line) is ignored
         and is not used for sorting purposes.

"""

import sys
from operator import attrgetter
from collections import defaultdict

g_program_name = __file__.split('/')[-1]
g_version_str = '0.12.0'
g_date_str = '2016-12-21'

# In order to specify any amino acid in a PDB file, you must provide 3
# identifiers:
#   the ChainID a single letter specifying
#   the SeqNum an integer indicating the location within that chain
#   the ICode ("insert code" usually 0.  I don't know why this number is
#              necessary. ..For the record, I never loved the PDB file format.)


class AtomDescr:

    def __init__(self, setChainID, setSeqNum, setICode, setAtomID):
        self.chainID = setChainID
        self.seqNum = setSeqNum
        self.iCode = setICode
        self.atomID = setAtomID

    # I plan to store this information in a python dictionary.
    # Unfortunately, in order to use such classes as keys in python dictionaries
    # I must define comparison operators, and a hash function.
    # In retrospect I figured out it would have been easier just to use tuples
    # as dictionary keys.  I suppose it was a good excercise to try to do it
    # using python classes instead.  I'm sorry it made the code so long.

    def __le__(self, other):
        # return ((self.chainID < other.chainID) or ((self.chainID == other.cha
        # instead I'll exploit python's ability to compare tuples
        return (self.chainID, self.seqNum, self.iCode, self.atomID) <= (other.chainID, other.seqNum, other.iCode, other.atomID)

    def __lt__(self, other):
        return (self.chainID, self.seqNum, self.iCode, self.atomID) < (other.chainID, other.seqNum, other.iCode, other.atomID)

    def __eq__(self, other):
        # return ((self.chainID == x.chainID) and (self.seqNum == x.seqNum) and
        # (self.iCode == x.iCode))
        return (self.chainID, self.seqNum, self.iCode, self.atomID) == (other.chainID, other.seqNum, other.iCode, other.atomID)

    def __ne__(self, other):
        return not __eq__(self, other)

    def __gt__(self, other):
        return not __le__(self, other)

    def __ge__(self, other):
        return not __lt__(self, other)

    def __cmp__(self, other):
        if __lt__(self, other):
            return -1
        elif __gt__(self, other):
            return 1
        else:
            return 0

    def __hash__(self):
        numChainIDs = 128
        numICodes = 128
        i = self.seqNum
        i *= numChainIDs
        i += ord(self.chainID)
        i *= numICodes
        i += ord(self.iCode)
        i *= 10
        i += self.atomID
        return i


def main():
    if len(sys.argv) == 1:
        use_all_residues = True
    elif len(sys.argv) == 7:
        use_all_residues = False
        first = AtomDescr(sys.argv[1], int(sys.argv[2]), sys.argv[3], 0)
        last = AtomDescr(sys.argv[4], int(sys.argv[5]), sys.argv[6], 2147483647)
    else:
        sys.stderr.write("Error(" + g_program_name + "): This program requires either 0 or 6 arguments.\n"
                         "       By default, the the sequence is extracted from the entire PDB file.\n"
                         "       In that case, no arguments are required.\n"
                         "       Alternately, you can limit the selection to a single interval of\n"
                         "       residues from one of the chains in the PDB file.\n"
                         "       To specify an interval, you must passing 6 arguments to this program.\n"
                         "       This program requires a pair of residues to designate the first and\n"
                         "       last members of the interval.  Each residue requires 3 identifiers.\n"
                         "       Consequently the six arguments needed are:\n"
                         "ChainID_first SeqNum_first ICode_first ChainID_last SeqNum_last ICode_last\n")
        exit(-1)


    atoms2lines = defaultdict(list)

    for line in sys.stdin:
        if (line[0:6] == "ATOM  ") or (line[0:6] == "HETATM"):
            atomID = int(line[6:11])
            #atomType  = line[12:16]
            #altLoc    = line[16:17]
            iCode = line[26:27]
            #resType   = line[17:20]
            chainID = line[21:22]
            seqNumStr = line[22:26]
            seqNum = int(seqNumStr)
            atomdescr = AtomDescr(chainID, int(seqNumStr), iCode, int(atomID))
            atoms2lines[atomdescr].append(line.rstrip('\n'))
        else:
            sys.stdout.write(line)

    # Extract an (unordered) list of the atomdescrs of the atoms in the sequence
    atomdescrs = [atomdescr for atomdescr in atoms2lines]

    # Residues in PDB files are often not listed in order.
    # Consequently, we must sort the list by chainID, seqNum, and finnaly iCode:
    sequence_of_atomdescrs = sorted(atomdescrs, key=attrgetter(
        'chainID', 'seqNum', 'iCode', 'atomID'))

    for atomdescr in sequence_of_atomdescrs:
        for line in atoms2lines[atomdescr]:
            sys.stdout.write(line + '\n')

    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2020, Scripps Research
# All rights reserved.

man_page_text = """
Usage (example):

postprocess_transitions.py ttree_assignments.txt -t file.template \
    > new_file.template

Where "file.template" contains contents similar to:

if atoms == [@atom:A,@atom:B,@atom:C] and
   bonds == [[1,2], [2,3]] and
   bonding_ids == [1,2] and
   edge_ids == [3]
then
   atoms = [@atom:B,@atom:B,@atom:C] and
   bonds = [[2,3], [3,1]]

Eventually, I will also support this syntax:

if atoms @atom:A @atom:B* @atom:C and 
   bond_type[1,2] == @bond:AB and
   bond_type[2,3] == @bond:BC and
   angle_type[1,2,3] == @angle:ABC and
   distance[1,3] < 3.0 and prob(0.5)
then
   atom_type[2] = @atom:B and
   atom_type[3] = @atom:D and
   delete_bond(1,2) and
   bond_type[1,3] = @bond:AD

if atom_type[1] == @atom:A and 
   atom_type[2] == @atom:B* and
   atom_type[3] == @atom:C and
   bond_type[1,2] == @bond:AB and
   bond_type[2,3] == @bond:BC and
   angle_type[1,2,3] == @angle:ABC and
   distance[1,3] < 3.0 and prob(0.5)
then
   atom_type[2] = @atom:B and
   atom_type[3] = @atom:D and
   delete_bond(1,2) and
   bond_type[1,3] = @bond:AD

if atom_type[1] == @atom:A and 
   atom_type[2] == @atom:B* and
   atom_type[3] == @atom:C and
   bond_type[1,2] == @bond:AB and
   bond_type[2,3] == @bond:BC and
   rmsd((1,2,3), ((1.3,5.2,1.2),(2.4,4.5,6.6),(0.01,1.5,9.55)) <= 3.2
   then
   delete_bond(2,3)
   delete_atom(3)

if atom_type[1] == @atom:A and 
   atom_type[2] == @atom:B* and
   atom_type[3] == @atom:C and
   bond_type[1,2] == @bond:AB and
   bond_type[2,3] == @bond:BC and
   rmsd((1,2,3), ((1.3,5.2,1.2),(2.4,4.5,6.6),(0.01,1.5,9.55)) <= 3.2
then
   coords = ((1.3,5.2,1.2),(2.4,4.5,6.6),(0.01,1.5,9.55),(-1.2,0.1,12)) #add atom
   and atom_type[4] = @atom:D
   and bond_type[3,4] = @bond:CD

"""


# Keywords recognized by this script:
# '[', ']', '(', ')', ',', '==', '=', 'atom_type', 'bond_type', 'angle_type', 'dihedral_type', 'improper_type', 'distance', 'prob', 'rmsd', 'coords', 'delete_atom', 'delete_bond', 'delete_angle', 'delete_dihedral', 'dihedral_improper'
#
# ------ the following features not needed for version 1.0: ------
#
#1) This next step is only needed for rmsd((),())  and coords():
#   Create a variant of SplitQuotedString() that splits according to both
#   parenthesis and commas and works with nested expressions.
#   Call it "SplitNestedQuotes()".  It accepts these arguments with these
#   default values:
#      delim = ','
#      open_paren = '(',
#      close_paren = ')',
#   It will split template lists of this form
#    ['bond((1,2),', VarRef('@/bond:AB'), '), ((2,3),', VarRef('@/bond:BC'),'))']
#        ... which is what ReadTemplate() will return when invoked on
#            'bond(((1,2),@/bond:AB), ((2,3),@/bond:BC))'
#   into something like this:
#        KRUFT ALERT.  THE NEXT FEW LINES OF COMMENTS ARE OUT OF DATE -AJ2020-11
#    ['bond',
#     '(',
#     ['(1,2),', VarRef('@/bond:AB')],
#     ['(2,3),', VarRef('@/bond:BC')],
#     ')']
#   Note: This function only processes the outermost paren expression.
#         The '(1,2),' was not processed further.  Had it been, it would have
#         returned [['(', ['1','2'] ,')'], '']
#   
#2) Use SplitNestedQuotes() to find the arguments following the
#   rmsd() and coords() keywords.

import sys
import argparse
from collections import defaultdict
import re
import gc

try:
    from .ttree import ExtractFormattingCommands
    from .ttree_lex import *

except (ImportError, SystemError, ValueError):
    # not installed as a package
    from ttree import ExtractFormattingCommands
    from ttree_lex import *


g_filename = __file__.split('/')[-1]
g_module_name = g_filename
if g_filename.rfind('.py') != -1:
    g_module_name = g_filename[:g_filename.rfind('.py')]
g_date_str = '2020-11-04'
g_version_str = '0.0.3'
g_program_name = g_filename
#sys.stderr.write(g_program_name+' v'+g_version_str+' '+g_date_str+' ')





def main():
    try:
        ap = argparse.ArgumentParser(prog=g_program_name)
        ap.add_argument('bindings_filename',  # (positional argument)
                        help='assignments file name (usually "ttree_assignments.txt")')
        ap.add_argument('-t', '--template', dest='template', required=False,
                        help='template text file (typically generated by moltemplate, and ending in ".template")')
        args = ap.parse_args()
        bindings_filename = args.bindings_filename
        f = open(bindings_filename)

        atom_types = set([])
        bond_types = set([])
        angle_types = set([])
        dihedral_types = set([])
        improper_types = set([])

        # The line above is robust but it uses far too much memory.
        # This for loop below works for most cases.
        for line in f:
            #tokens = lines.strip().split()
            # like split but handles quotes
            tokens = SplitQuotedString(line.strip())
            if len(tokens) < 2:
                continue
            if tokens[0].find('@') != 0:
                continue
            if tokens[0][2:].find('atom') == 0:
                atom_types.add(tokens[0][1:])
            elif tokens[0][2:].find('bond') == 0:
                bond_types.add(tokens[0][1:])
            elif tokens[0][2:].find('angle') == 0:
                angle_types.add(tokens[0][1:])
            elif tokens[0][2:].find('dihedral') == 0:
                dihedral_types.add(tokens[0][1:])
            elif tokens[0][2:].find('improper') == 0:
                improper_types.add(tokens[0][1:])

        f.close()
        gc.collect()


        # Now open the template file containing the list of transition rules.

        if args.template is not None:
            templ_file = open(args.template, 'r')
            lex = LineLex(templ_file)
        else:
            templ_file = sys.stdin
            lex = TtreeShlex(sys.stdin,
                             '__standard_input_for_postprocess_coeffs__')
        lex.commenters = '#'           #(don't attempt to skip over comments)
        lex.line_extend_chars += '&'   #(because LAMMPS interprets '&' as '\')

        transition_rules_orig = []
        if_clause = []
        then_clause = []
        in_if_clause = True
        while lex:
            token = lex.get_token()
            if ((token == '') or (token == 'if')):
                if ((len(if_clause)>0) and (len(then_clause)>0)):
                    transition_rules_orig.append([if_clause, then_clause])
                    if_clause = []
                    then_clause = []
                if token == 'if':
                    in_if_clause = True
                    continue
                elif token == '':
                    break
            elif token == 'then':
                then_clause = []
                in_if_clause = False
                continue
            elif token in ('@', '$'):
                var_name = GetVarName(lex)
                token = token + var_name  # for example: "@/atom:GAFF2/c3"
            if in_if_clause:
                if_clause.append(token)
            else:
                then_clause.append(token)

        # now close the file (if we opened it)
        if args.template is not None:
            templ_file.close()

        # Now split the if and then clauses into tokens separated by "and"
        if_requirements = []
        if_requirement = []
        then_results = []
        then_result = []
        for rule in transition_rules_orig:
            if_clause = rule[0]
            then_clause = rule[1]
            #split the if_clause into lists of tokens separated by 'and':
            for token in if_clause:
                if ((token == 'and') and (len(if_requirement) > 0)):
                    if_requirements.append(if_requirement)
                    if_requirement = []
                else:
                    if_requirement.append(token)
            if len(if_requirement) > 0:
                if_requirements.append(if_requirement)
            # Replace rule[0] with the if_requirements list
            rule[0] = if_requirements

            #split the then_clause into lists of tokens separated by 'and'
            for token in then_clause:
                if ((token == 'and') and (len(then_result) > 0)):
                    then_results.append(then_result)
                    then_result = []
                else:
                    then_result.append(token)
            if len(then_result) > 0:
                then_results.append(then_result)
            # Replace rule[1] with the then_results list
            rule[1] = then_results


        # Now loop through all of the transition rules.  For each rule,
        # figure out how many times the user specified an atom type, or
        # bonded type, or angle type or dihedral type or improper type
        # which must be satisfied in order to satisfy the if conditions.
        #
        # Then, for that rule, generate a separate transition rule for
        # every possible combination of atom types, bond types, angle types,
        # dihedral types, and improper types which satisfies the requirements
        # specified by the user after considering wildcards and regex.
        # In this way, a single rule specified by the user (with vague atom
        # type names or vague bonded type napes) might get translated
        # (expanded) into a large number of individual transition rules
        # for use with fix bond/react, where in each rule, each atom type,
        # bond type, angle type, etc... is specified explicitly.

        Natoms = 0 # we will store the number of atoms in the
                   # pre-reaction template here
        transition_rules = [] # we will store processed transition rules here
        atom_req = []  # atom type requirements
        bond_req = []  # bond type requirements
        angle_req = []  # angle type requirements
        dihedral_req = []  # dihedral type requirements
        improper_req = []  # improper type requirements
        for rule in transition_rules_orig:
            if_requirements = rule[0]
            for if_requirement in if_requirements:
                tokens = if_requirement
                assert(len(tokens) > 0)
                iatm = -1
                if tokens[0] == 'atom':
                    # allow users to use either '=' or '==' to test for equality
                    # For example, these should both work:
                    #    'if atom[1] == @atom:A'
                    #    'if atom[1] = @atom:A'.
                    # But if '==' was used, the lexer will mistake this for
                    # two separate tokens ('=' followed by '=').  We take care
                    # of that here by deleting the extra '='
                    if (tokens[4] == tokens[5] == '='):
                        tokens[4] = '=='
                        del tokens[5]
                    if not ((len(tokens) == 6) and
                            (tokens[1] == '[') and
                            (tokens[2].isnumeric() and
                             (int(tokens[2]) > 0)) and
                            (tokens[3] == ']') and
                            (tokens[4] == '==') and
                            ((tokens[5].find('@') != -1) and
                             (tokens[5].find('atom:') != -1))):
                        raise InputError('Error in transitions file near:\n'+
                                         '      '+' '.join(tokens)+'\n')

                iatm = int(tokens[2])
                if iatm >= Natoms:
                    atom_req += [[] for i in range(0, 1 + iatm - Natoms)]
                    Natoms = iatm + 1
                    assert(Natoms == len(atom_req))

                typestr = tokens[5][1:]  # a string identifying atom type(s)
                                         # eg: '@atom:/SPCE/O' or '@atom:C*'
                #icolon = typestr.find('atom:') + 5
                #typestr = typestr[icolon:]

                # If the second token is surrounded by '/' characters, interpret
                # it as a regular expression.
                type_is_re = HasRE(typestr)

                # If the second token contains wildcard characters, interpret
                # it as a wildcard (ie. glob) expression.
                type_is_wild = (HasWildcard(typestr) #does it contain *,?
                                and
                                (typestr[0] != '{')) #(ignore * or ? in {})

                if type_is_re:
                    regex_str = VarNameToRegex(typestr)
                    typepattern = re.compile(regex_str)
                else:
                    #left_paren, typepattern, text_after=ExtractVarName(typestr)
                    typepattern = typestr
                    left_paren = ''
                    text_after = ''
                for atype in atom_types:
                    if MatchesPattern(atype, typepattern):
                        #atom_req[iatm].append('@'+left_paren+atype+text_after)
                        atom_req[iatm].append('@{'+atype+'}')

        # ------------------ CONTINUEHERE --------------------
                    


    except (ValueError, InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(-1)

    return

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013

"""
   renumber the integers at the beginning of ever line in the file
   to make sure these numbers are contiguous.
   The file is read from sys.stdin.
   This program does not parse an entire LAMMPS data file.
   The text from the "Atoms" section
    (or "Bonds", or "Angles", or "Dihedrals", or "Impropers" sections)
   of the LAMMPS file must be extracted in advance.

"""

import sys
from operator import itemgetter

def main():
    in_stream = sys.stdin
    f = None
    fname = None
    if len(sys.argv) == 2:
        fname = sys.argv[1]
        f = open(fname, 'r')
        in_stream = f


    lines = in_stream.readlines()
    column1_iorig_columnsAfter1 = []

    # Start at the end of the file and read backwards.
    # If duplicate lines exist, eliminate the ones that occur earlier in the file.
    i = 0
    while i < len(lines):
        line_orig = lines[i]
        line = line_orig.rstrip('\n')
        comment = ''
        if '#' in line_orig:
            ic = line.find('#')
            line = line_orig[:ic]
            comment = ' ' + line_orig[ic:].rstrip('\n')

        tokens = line.strip().split()
        if len(tokens) > 0:
            if str.isdigit(tokens[0]):
                column1 = int(tokens[0])
            else:
                column1 = tokens[0]
            column1_iorig_columnsAfter1.append(
                [column1, i, ' '.join(tokens[1:]) + comment])
        i += 1

    # Sort the list of lines by the first number on each line

    column1_iorig_columnsAfter1.sort(key=itemgetter(0))

    # Change all of these numbers so that they are consecutive starting at 1
    for i in range(0, len(column1_iorig_columnsAfter1)):
        column1_iorig_columnsAfter1[i][0] = i + 1

    # Sort the list of lines so they are back in the original order

    column1_iorig_columnsAfter1.sort(key=itemgetter(1))

    for i in range(0, len(column1_iorig_columnsAfter1)):
        column1 = column1_iorig_columnsAfter1[i][0]
        columnsAfter1 = column1_iorig_columnsAfter1[i][2]
        sys.stdout.write(str(column1) + ' ' + columnsAfter1 + '\n')

    if f != None:
        f.close()

    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013

"""
   Reorder the atoms in the Angles section of a data file to make sure that
   atoms have a "canonical order" (for example the first atom has a lower
   id than the last atom, for angle and dihedral interactions.
   (This helps us detect potential problems like dupicate Angle interactions.)

"""

from operator import itemgetter
import importlib
import os
import sys
sys.path.append(os.getcwd())

g_program_name = __file__.split('/')[-1]


def main():
    in_stream = sys.stdin

    section_name = ''
    if len(sys.argv) == 3:
        section_name = sys.argv[1]
        bond_pattern_module_name = sys.argv[2]
        # If the file name ends in ".py", then strip off this suffix.
        # The next line does not work. Too lazy do care why.
        # bond_pattern_module_name=bond_pattern_module_name.rstrip('.py')
        # Do this instead
        pc = bond_pattern_module_name.rfind('.py')
        if pc != -1:
            bond_pattern_module_name = bond_pattern_module_name[0:pc]

    else:
        sys.stderr.write('Usage Example:\n\n'
                         '      ' + g_program_name + ' Angles nbody_angles.py < angles.txt > new_angles.txt\n\n'
                         '      In this example \"angles.txt\" contains only the \"Angles\" section of\n'
                         '      a LAMMPS DATA file.  (Either a text-editor, or the \n'
                         '      \"extract_lammps_data.py\" script can be used to select a section from\n'
                         '      a LAMMPS DATA file\n\n'
                         'Error(' + g_program_name +
                         '): expected exactly one argument:\n'
                         '       \"Angles\",  \"Dihedrals\", or \"Impropers\"\n')
        exit(-1)

    # Ordering rules are defined in a seperate module named
    # nbody_angles.py, nbody_dihedrals.py, nbody_impropers.py
    # Load that now.

    # search locations
    package_opts = [[bond_pattern_module_name, __package__],
                    ['nbody_alt_symmetry.'+bond_pattern_module_name,
                     __package__]]

    if __package__:
        for i in range(0, len(package_opts)):
            package_opts[i][0] = '.' + package_opts[i][0]

    grph = None
    for name, pkg in package_opts:
        try:
            # define grph.bond_pattern, grph.canonical_order
            grph = importlib.import_module(name, pkg)
            break
        except (ImportError, SystemError, ValueError):
            pass

    if grph is None:
        sys.stderr.write('Error: Unable to locate file \"' +
                         bond_pattern_module_name + '\"\n'
                         '       (Did you mispell the file name?\n'
                         '        Check the \"nbody_alt_symmetry/\" directory.)\n')
        sys.exit(-1)

    # This module defines the graph representing the bond pattern for this type
    # of interaction.  (The number of vertices and edges for the graph corresponds
    # to the number of atoms and bonds in this type of interaction.)
    natoms = grph.bond_pattern.GetNumVerts()
    nbonds = grph.bond_pattern.GetNumEdges()


    for line_orig in in_stream:
        line = line_orig.rstrip('\n')
        comment = ''
        if '#' in line_orig:
            ic = line.find('#')
            line = line_orig[:ic]
            comment = ' ' + line_orig[ic:].rstrip('\n')

        tokens = line.strip().split()
        swapped = False
        if len(tokens) == 2 + natoms:
            all_integers = True
            abids_l = [[0 for i in range(0, natoms)],
                       [0 for i in range(0, nbonds)]]
            for i in range(0, natoms):
                if not tokens[2 + i].isdigit():
                    all_integers = False
            if all_integers:
                for i in range(0, natoms):
                    abids_l[0][i] = int(tokens[2 + i])
            else:
                for i in range(0, natoms):
                    abids_l[0][i] = tokens[2 + i]
            abids = grph.canonical_order((tuple(abids_l[0]), tuple(abids_l[1])))
            for i in range(0, natoms):
                tokens[2 + i] = str(abids[0][i])

        sys.stdout.write(' '.join(tokens) + comment + '\n')

    return


if __name__ == '__main__':
    main()



from .ttree import BasicUISettings, BasicUIParseArgs, EraseTemplateFiles, \
    StackableCommand, PopCommand, PopRightCommand, PopLeftCommand, \
    PushCommand, PushLeftCommand, PushRightCommand, ScopeCommand, \
    WriteVarBindingsFile, StaticObj, InstanceObj, ExtractFormattingCommands, \
    BasicUI, ScopeBegin, ScopeEnd, WriteFileCommand, Render

from .ttree_lex import TtreeShlex, split, LineLex, SplitQuotedString, \
    EscCharStrToChar, SafelyEncodeString, RemoveOuterQuotes, MaxLenStr, \
    HasWildcard, HasRE, InputError, ErrorLeader, SrcLoc, OSrcLoc, TextBlock, \
    VarRef, VarNPtr, VarBinding, SplitTemplate, SplitTemplateMulti, \
    TableFromTemplate, ExtractCatName, DeleteLinesWithBadVars, TemplateLexer

from .nbody_graph_search import Disconnected, NotUndirected, Edge, Vertex, \
     Dgraph, Ugraph, SortVertsByDegree, DFS, GraphMatcher 

from .nbody_by_type_lib import GenInteractions_int, GenInteractions_str

from .lttree import LttreeSettings, LttreeParseArgs, TransformAtomText, \
    TransformEllipsoidText, AddAtomTypeComments, ExecCommands, WriteFiles

from .lttree_styles import AtomStyle2ColNames, ColNames2AidAtypeMolid, \
    ColNames2Coords, ColNames2Vects, ColNames2Vects, data_atoms, data_masses

from .ettree_styles import \
    LinesWSlashes, SplitMultiDelims, SplitAtomLine, \
    iEsptAtomCoords, iEsptAtomVects, iEsptAtomType, iEsptAtomID

from .ttree_matrix_stack import MultMat, MatToStr, LinTransform, \
    AffineTransform, AffineCompose, CopyMat, ScaleMat, RotMatAXYZ, \
    CrossProd, DotProd, Length, Normalize, RotMatXYZXYZ, MultiAffineStack

# Stand-alone executable scripts
from .ttree import main
from .lttree import main
from .ettree import main
from .ttree_render import main
from .ltemplify import main, Ltemplify
from .dump2data import main
from .raw2data import main
from .extract_lammps_data import main
from .genpoly_lt import main, GenPoly, GPSettings
from .genpoly_modify_lt import main, GenPolyMod, GPModSettings, DistributePeriodic, DistributeRandom
from .interpolate_curve import main, ResampleCurve, CalcNaturalCubicSplineCoeffs, SplineEval, SplineEvalD1, SplineEvalD2, SplineInterpEval, SplineInterpEvalD1, SplineInterpEvalD2, SplineCurvature2D, SplineInterpCurvature2D
from .nbody_by_type import main

__all__ = [# General modules for parsing and rendering text templates:
           'ttree','ttree_lex','ttree_render',
           # General modules for handling force-fields:
           'nbody_graph_search','nbody_by_type_lib','nbody_by_type',
           'nbody_Angles','nbody_Bonds','nbody_Dihedrals','nbody_Impropers',
           'nbody_reorder_atoms',
           # Coordinate transformations:
           'ttree_matrix_stack',
           'recenter_coords',
           'genpoly_lt',
           'interpolate_curve',
           'pdbsort',
           # LAMMPS specific:
           'lttree','lttree_styles','lttree_check','lttree_postprocess',
           'dump2data', 'raw2data',
           'extract_lammps_data',
           'ltemplify',
           'postprocess_coeffs','postprocess_input_script',
           'renumber_DATA_first_column',
           'remove_duplicate_atoms','remove_duplicates_nbody',
           'bonds_by_type','charge_by_bond',
           # ESPResSo specific:
           'ettree','ettree_styles','extract_espresso_atom_types']



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013

"""
   Get rid of lines containing duplicate bonded nbody interactions in the
   corresponding section of a LAMMPS data file (such as bonds, angles,
   dihedrals and impropers).  Duplicate lines which occur later are
   preserved and the earlier lines are erased.
   (This program reads from sys.stdin.  This program does not parse the entire
    data file.  The text from the relevant section of the LAMMPS file should be
   extracted in advance before it is sent to this program.)

"""

import sys

def main():
    in_stream = sys.stdin

    if len(sys.argv) == 2:
        n = int(sys.argv[1])
    if (len(sys.argv) != 2) or (n < 1):
        sys.stderr.write(
            'Error (remove_duplicates_nbody.py): expected a positive integer argument.\n')
        sys.exit(-1)

    atom_ids_in_use = set([])

    lines = in_stream.readlines()

    # Start at the end of the file and read backwards.
    # If duplicate lines exist, eliminate the ones that occur earlier in the file.
    i = len(lines)
    while i > 0:
        i -= 1
        line_orig = lines[i]
        line = line_orig.rstrip('\n')
        if '#' in line_orig:
            ic = line.find('#')
            line = line_orig[:ic]

        tokens = line.strip().split()
        if len(tokens) == 2 + n:
            atom_ids = tuple(tokens[2:2 + n])
            if atom_ids in atom_ids_in_use:
                del lines[i]
            else:
                atom_ids_in_use.add(atom_ids)
        elif len(tokens) == 0:
            del lines[i]

    for line in lines:
        sys.stdout.write(line)

    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2017, California Institute of Technology
# All rights reserved.

"""
   Reorder the integer arguments to the commands in a LAMMPS input
   file if these arguments violate LAMMPS order requirements.
   We have to do this because the moltemplate.sh script will automatically
   assign these integers in a way which may violate these restrictions
   and the user has little control over this.

   This script:
   swaps the I and J integers in    "pair_coeff I J ..." commands when I > J

   Other features may be added later

"""


import sys

def main():
    lines_orig = []
    f = None
    fname = None
    num_lines_ignore = 0


    # Lines from files passed as arguments are read and processed silently.
    # (Why? Sometimes it's necessary to read the contents of previous input scripts
    #  in order to be able to understand a script command which appears later.
    #  I'm assuming these files will be processed by lammps in the same order. So I
    #  must insure that moltemplate.sh passes them to this program in that order.
    #  I'm too lazy to read the "include" commands in input scripts correctly.)
    if len(sys.argv) > 1:
        for fname in sys.argv[1:]:
            f = open(fname, 'r')
            in_stream = f
            lines_orig += in_stream.readlines()
            num_lines_ignore += len(lines_orig)
            f.close()


    # Lines read from the standard input are read, processed, and printed to stdout
    in_stream = sys.stdin
    lines_orig += in_stream.readlines()

    pair_style_list = []
    swap_occured = False
    warn_wildcard = False

    i = 0
    while i < len(lines_orig):
        # Read the next logical line
        # Any lines ending in '&' should be merged with the next line before
        # breaking
        line_orig = ''
        while i < len(lines_orig):
            line_counter = 1 + i - num_lines_ignore
            line_orig += lines_orig[i]
            if ((len(line_orig) < 2) or (line_orig[-2:] != '&\n')):
                break
            i += 1
        line = line_orig.replace('&\n', '\n').rstrip('\n')

        comment = ''
        if '#' in line_orig:
            ic = line.find('#')
            line = line_orig[:ic]
            # keep track of comments (put them back later)
            comment = line_orig[ic:].rstrip()

        tokens = line.strip().split()
        if ((len(tokens) >= 2) and (tokens[0] == 'pair_style')):
            pair_style_list = tokens[1:]

        if ((len(tokens) >= 3) and (tokens[0] == 'pair_coeff')):

            if ((tokens[1].isdigit() and (tokens[2].isdigit())) and
                (int(tokens[1]) > int(tokens[2]))):

                swap_occured = True
                tmp = tokens[2]
                tokens[2] = tokens[1]
                tokens[1] = tmp

                if i >= num_lines_ignore:

                    # polite warning:
                    sys.stderr.write(
                        'swapped pair_coeff order on line ' + str(line_counter))
                    # if (fname != None):
                    #    sys.stderr.write(' of file \"'+fname+'\"')
                    sys.stderr.write('\n')

                    # Deal with the "hbond/" pair coeffs.
                    #
                    # The hbond/dreiding pair style designates one of the two atom types
                    # as a donor, and the other as an acceptor (using the 'i','j' flags)
                    # If swapped atom types eariler, we also need to swap 'i' with 'j'.
                    #
                    # If "hbond/dreiding.." pair style is used with "hybrid" or
                    # "hybrid/overlay" then tokens[3] is the name of the pair style
                    # and tokens[5] is either 'i' or 'j'.
                    if len(pair_style_list) > 0:
                        if ((pair_style_list[0] == 'hybrid') or
                                (pair_style_list[0] == 'hybrid/overlay')):
                            if ((len(tokens) > 5) and (tokens[5] == 'i') and (tokens[3][0:6] == 'hbond/')):
                                tokens[5] = 'j'
                                sys.stderr.write(
                                    '  (and replaced \"i\" with \"j\")\n')
                            elif ((len(tokens) > 5) and (tokens[5] == 'j') and (tokens[3][0:6] == 'hbond/')):
                                tokens[5] = 'i'
                                sys.stderr.write(
                                    '  (and replaced \"j\" with \"i\")\n')
                        elif (pair_style_list[0][0:6] == 'hbond/'):
                            if ((len(tokens) > 4) and (tokens[4] == 'i')):
                                tokens[4] = 'j'
                                sys.stderr.write(
                                    '  (and replaced \"i\" with \"j\")\n')
                            elif ((len(tokens) > 4) and (tokens[4] == 'j')):
                                tokens[4] = 'i'
                                sys.stderr.write(
                                    '  (and replaced \"j\" with \"i\")\n')

                    sys.stdout.write(
                        (' '.join(tokens) + comment).replace('\n', '&\n') + '\n')

            else:
                if ((('*' in tokens[1]) or ('*' in tokens[2]))
                    and
                    (not (('*' == tokens[1]) and ('*' == tokens[2])))):
                    warn_wildcard = True
                if i >= num_lines_ignore:
                    sys.stdout.write(line_orig)
        else:
            if i >= num_lines_ignore:
                sys.stdout.write(line_orig)

        i += 1


    if swap_occured:
        sys.stderr.write('\n'
                         '  WARNING: Atom order in some pair_coeff commands was swapped to pacify LAMMPS.\n'
                         '  For some exotic pair_styles such as hbond/dreiding, this is not enough. If you\n'
                         '  use exotic pair_styles, please verify the \"pair_coeff\" commands are correct.\n')

    if warn_wildcard:
        sys.stderr.write('\n'
                         '  WARNING: The use of wildcard characters (\"*\") in your \"pair_coeff\"\n'
                         '           commands is not recommended.\n'
                         '           (It is safer to specify each interaction pair manually.\n'
                         '            Check every pair_coeff command.  Make sure that every atom type in\n'
                         '            the first group is <= atom types in the second group.\n'
                         '            Moltemplate does NOT do this when wildcards are used.)\n'
                         '        If you are using a many-body pair style then ignore this warning.\n')

    return

if __name__ == '__main__':
    main()



# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013


try:
    from .nbody_graph_search import Ugraph
except (ImportError, SystemError, ValueError):
    # not installed as a package
    from nbody_graph_search import Ugraph

# This file defines how dihedral interactions are generated by moltemplate.sh
# by default.  It can be overridden by supplying your own custom file.

#    To find 4-body "dihedral" interactions, we would use this subgraph:
#
#                              1st bond connects atoms 0 and 1
#       *---*---*---*      =>  2nd bond connects atoms 1 and 2
#       0   1   2   3          3rd bond connects atoms 2 and 3
#

bond_pattern = Ugraph([(0, 1), (1, 2), (2, 3)])
# (Ugraph atom indices begin at 0, not 1)


def canonical_order(match):
    """
    Before defining a new interaction, we must check to see if an
    interaction between these same 4 atoms has already been created
     (perhaps listed in a different, but equivalent order).
    If we don't check for this this, we will create many unnecessary redundant
    interactions (which can slow down he simulation).
    To avoid this, I define a "canonical_order" function which sorts the atoms
    and bonds in a way which is consistent with the symmetry of the interaction
    being generated...  Later the re-ordered list of atom and bond ids will be
    tested against the list of atom/bond ids in the matches-found-so-far,
    before it is added to the list of interactions found so far.  Note that
    the energy of a dihedral interaction is a function of the dihedral-angle.
    The dihedral-angle is usually defined as the angle between planes formed
    by atoms 0,1,2 & 1,2,3. This angle does not change when reversing the
    order of the atoms.  So it does not make sense to define a separate
    dihedral interaction between atoms  0,1,2,3   AS WELL AS between   3,2,1,0.
    So we sort the atoms so that the first atom has a lower atomID than the
    last atom.  (Later we will check to see if we have already defined an
    interaction between these 4 atoms.  If not then we create a new one.)

    """

    # match[0][0:3] contains the ID numbers of the 4 atoms in the match
    atom0 = match[0][0]
    atom1 = match[0][1]
    atom2 = match[0][2]
    atom3 = match[0][3]
    # match[1][0:2] contains the ID numbers of the the 3 bonds
    bond0 = match[1][0]
    bond1 = match[1][1]
    bond2 = match[1][2]
    if atom0 < atom3:
        # return ((atom0, atom1, atom2, atom3), (bond0, bond1, bond2))  same
        # as:
        return match
    else:
        return ((atom3, atom2, atom1, atom0), (bond2, bond1, bond0))



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2014

"""
extract_lammps_data.py is a simple script which extracts sections of text from
a LAMMPS data file.

Typical usage: 

extract_lammps_data.py SECTION_NAME < FILE.DATA > EXCERPT.TXT

This extracts a section from a LAMMPS data file and saves it in EXCERPT.TXT.

More general usage:

extract_lammps_data.py [-n] SECTION_LIST < FILE.DATA > EXCERPT.TXT

For more details, see "doc/utils/docs_extract_lammps_data.txt"
"""

import sys

lammps_data_sections = set(['Atoms',
                            'Masses',
                            'Bonds',
                            'Bond Coeffs',
                            'Angles',
                            'Angle Coeffs',
                            'Dihedrals',
                            'Dihedral Coeffs',
                            'Impropers',
                            'Improper Coeffs',
                            'BondBond Coeffs',          # class2 angles
                            'BondAngle Coeffs',         # class2 angles
                            'MiddleBondTorsion Coeffs',  # class2 dihedrals
                            'EndBondTorsion Coeffs',    # class2 dihedrals
                            'AngleTorsion Coeffs',      # class2 dihedrals
                            'AngleAngleTorsion Coeffs',  # class2 dihedrals
                            'BondBond13 Coeffs',        # class2 dihedrals
                            'AngleAngle Coeffs',        # class2 impropers
                            'Angles By Type',   # new. not standard LAMMPS
                            'Dihedrals By Type',  # new. not standard LAMMPS
                            'Angles By Type'])   # new. not standard LAMMPS


def DeleteComments(string,
                   escape='\\',
                   comment_char='#'):
    escaped_state = False
    for i in range(0, len(string)):
        if string[i] in escape:
            if escaped_state:
                escaped_state = False
            else:
                escaped_state = True
        elif string[i] == comment_char:
            if not escaped_state:
                return string[0:i]
    return string


def ExtractDataSection(f,
                       section_name,
                       comment_char='#',
                       include_section_name=False,
                       return_line_nums=False):

    inside_section = False
    if section_name in ('header', 'Header'):  # "Header" section includes beginning
        inside_section = True

    nonblank_encountered = False
    nonheader_encountered = False

    i = 0
    for line_orig in f:
        return_this_line = False
        line = DeleteComments(line_orig).strip()
        if line in lammps_data_sections:
            nonheader_encountered = True
        if section_name in ('header', 'Header'):
            # The "header" section includes all lines at the beginning of the
            # before any other section is encountered.
            if nonheader_encountered:
                return_this_line = False
            else:
                return_this_line = True
        elif line == section_name:
            inside_section = True
            nonblank_encountered = False
            if include_section_name:
                return_this_line = True
        # A block of blank lines (which dont immediately follow
        # the section_name) signal the end of a section:
        elif len(line) == 0:
            if inside_section and include_section_name:
                return_this_line = True
            if nonblank_encountered:
                inside_section = False
        elif line[0] != comment_char:
            if inside_section:
                nonblank_encountered = True
                return_this_line = True

        if return_this_line:
            if return_line_nums:
                yield i
            else:
                yield line_orig

        i += 1


def main():
    lines = sys.stdin.readlines()
    exclude_sections = False
    if sys.argv[1] == '-n':
        exclude_sections = True
        del sys.argv[1]

    if not exclude_sections:
        for section_name in sys.argv[1:]:
            for line in ExtractDataSection(lines, section_name):
                sys.stdout.write(line)
    else:
        line_nums_exclude = set([])
        for section_name in sys.argv[1:]:
            for line_num in ExtractDataSection(lines,
                                               section_name,
                                               include_section_name=True,
                                               return_line_nums=True):
                line_nums_exclude.add(line_num)
        for i in range(0, len(lines)):
            if i not in line_nums_exclude:
                sys.stdout.write(lines[i])

    return

if __name__ == "__main__":
    main()
